@@ -11,12 +11,15 @@ d - dimension
 l - logits (text)
 i, j - sequence (row, col)
 p - positions
+s - residual streams
 """
 
 import os
 import math
 from collections import defaultdict
 
+from random import randrange
+from itertools import count
 from functools import partial, wraps, cache
 from typing import NamedTuple, Callable, Literal
 
@@ -37,8 +40,12 @@ from einops import rearrange, repeat, reduce, einsum, pack, unpack
 
 from ema_pytorch import EMA
 
+from axial_positional_embedding import ContinuousAxialPositionalEmbedding
+
 from rotary_embedding_torch import RotaryEmbedding, apply_rotary_emb
 
+from hyper_connections import HyperConnections
+
 from tqdm import tqdm
 from loguru import logger
 
@@ -118,6 +125,11 @@ def default(v, d):
 def identity(t):
     return t
 
+def always(val):
+    def inner(*args, **kwargs):
+        return val
+    return inner
+
 def first(it):
     return it[0]
 
@@ -142,6 +154,15 @@ def add_temp_batch_dim(fn: Callable):
         return out
     return inner
 
+def pack_with_inverse(t, pattern):
+    packed, packed_shape = pack(t, pattern)
+
+    def inverse(out, inv_pattern = None):
+        inv_pattern = default(inv_pattern, pattern)
+        return unpack(out, packed_shape, inv_pattern)
+
+    return packed, inverse
+
 def pack_one_with_inverse(t, pattern):
     packed, packed_shape = pack([t], pattern)
 
@@ -500,61 +521,6 @@ def min_p_filter(logits, min_p = 0.1):
     limit = min_p * max_probs
     return torch.where(probs < limit, float('-inf'), logits)
 
-# MLP parameterized N-dimensional positions
-
-class MLPAxialPositions(Module):
-    def __init__(
-        self,
-        *,
-        num_dimensions, # 2 for images, 3 for video, etc etc
-        dim,
-        expand_factor = 2.
-    ):
-        super().__init__()
-        self.num_dimensions = num_dimensions
-        dim_hidden = int(dim * expand_factor)
-
-        self.mlp = nn.Sequential(
-            nn.Linear(num_dimensions, dim),
-            nn.SiLU(),
-            nn.Linear(dim, dim_hidden),
-            nn.SiLU(),
-            nn.Linear(dim_hidden, dim)
-        )
-
-        # tensor typing
-
-        self._d = dim
-
-    @property
-    def device(self):
-        return next(self.parameters()).device
-
-    @typecheck
-    def forward(
-        self,
-        modality_shape: Int['p'] | torch.Size,
-        flatten_dims = False
-    ) -> Float['... {self._d}']:
-
-        if isinstance(modality_shape, torch.Size):
-            modality_shape = tensor(modality_shape)
-
-        modality_shape = modality_shape.to(self.device)
-
-        assert len(modality_shape) == self.num_dimensions
-        dimensions = modality_shape.tolist()
-
-        grid = torch.meshgrid([torch.arange(dim_len, device = self.device) for dim_len in dimensions], indexing = 'ij')
-        axial_positions = stack(grid, dim = -1)
-
-        pos_emb = self.mlp(axial_positions.float())
-
-        if flatten_dims:
-            pos_emb = rearrange(pos_emb, '... d -> (...) d')
-
-        return pos_emb
-
 # random fourier embedding
 
 class RandomFourierEmbed(Module):
@@ -776,7 +742,7 @@ class Attention(Module):
             nn.Linear(dim, heads),
             nn.Sigmoid(),
             Rearrange('b n h -> b h n 1') # add head dimension
-        ) if learned_value_residual_mix else (lambda _: 0.5)
+        ) if learned_value_residual_mix else always(0.5)
 
         self.to_gates = nn.Sequential(
             nn.Linear(dim, heads, bias = False),
@@ -926,7 +892,8 @@ class Transformer(Module):
         ff_kwargs: dict = dict(),
         attn_laser = False,
         unet_skips = True,
-        use_flex_attn = False
+        use_flex_attn = False,
+        num_residual_streams = 4
     ):
         super().__init__()
         self.use_flex_attn = use_flex_attn
@@ -940,6 +907,18 @@ class Transformer(Module):
             nn.SiLU()
         )
 
+        # hyper connections
+
+        assert num_residual_streams > 0
+        is_hyper_connection = num_residual_streams > 1
+        self.num_residual_streams = num_residual_streams
+
+        counter = count()
+
+        init_residual_fn, self.expand_stream, self.reduce_stream = HyperConnections.get_init_and_expand_reduce_stream_functions(num_residual_streams, disable = num_residual_streams == 1)
+
+        # layers
+
         layers = ModuleList([])
 
         for ind in range(depth):
@@ -956,7 +935,10 @@ class Transformer(Module):
             attn = AdaptiveWrapper(attn, dim = dim, dim_cond = dim * 4)
             ff = AdaptiveWrapper(ff, dim = dim, dim_cond = dim * 4)
 
-            layers.append(ModuleList([skip_proj, attn, ff]))
+            attn_residual = init_residual_fn(dim = dim, layer_index = next(counter))
+            ff_residual = init_residual_fn(dim = dim, layer_index = next(counter))
+
+            layers.append(ModuleList([skip_proj, attn, attn_residual, ff, ff_residual]))
 
         self.layers = layers
         self.norm = RMSNorm(dim)
@@ -1046,6 +1028,10 @@ class Transformer(Module):
         cache = default(cache, (None,))
         iter_cache = iter(cache)
 
+        # expand input into multiple residual streams for maybe hyper connection
+
+        x = self.expand_stream(x)
+
         # transformer layers as usual, using mask from above
 
         skips = []
@@ -1055,7 +1041,7 @@ class Transformer(Module):
 
         depth = len(self.layers)
 
-        for ind, (skip_proj, attn, ff) in enumerate(self.layers):
+        for ind, (skip_proj, attn, attn_residual, ff, ff_residual) in enumerate(self.layers):
             layer = ind + 1
 
             # skip connection
@@ -1075,6 +1061,8 @@ class Transformer(Module):
 
             # attention and feedforward
 
+            x, add_attn_residual = attn_residual(x)
+
             (attn_out, attn_values), kv_cache = attn(
                 x,
                 rotary_emb = rotary_emb,
@@ -1090,8 +1078,17 @@ class Transformer(Module):
 
             new_cache.append(kv_cache)
 
-            x = attn_out + x
-            x = ff(x, **adaptive_kwargs) + x
+            x = add_attn_residual(attn_out)
+
+            x, add_ff_residual = ff_residual(x)
+
+            ff_out = ff(x, **adaptive_kwargs)
+
+            x = add_ff_residual(ff_out)
+
+        # reduce multiple residual streams for maybe hyper connection
+
+        x = self.reduce_stream(x)
 
         assert len(skips) == 0
 
@@ -1186,9 +1183,9 @@ class Transfusion(Module):
 
             assert exists(modality_ndim), '`modality_num_dim` must be set if you wish to automatically inject axial positional embeddings'
 
-            pos_generating_mlp = MLPAxialPositions(
+            pos_generating_mlp = ContinuousAxialPositionalEmbedding(
                 dim = dim,
-                num_dimensions = modality_ndim,
+                num_axial_dims = modality_ndim,
             )
 
             self.pos_emb_mlp.append(pos_generating_mlp)
@@ -1220,6 +1217,10 @@ class Transfusion(Module):
 
         self.fallback_to_default_shape_if_invalid = fallback_to_default_shape_if_invalid
 
+        # store number of text tokens
+
+        self.num_text_tokens = num_text_tokens
+
         # entire "sentence" start and end id
 
         num_text_special_ids = 2
@@ -1445,6 +1446,13 @@ class Transfusion(Module):
 
         device = self.device
 
+        # handle edge case where there are no text tokens
+
+        if self.num_text_tokens == 0:
+            logger.warning(f'you have `num_text_tokens` set to 0, so `sample` will be forwarded to `generate_modality_only(batch_size: int, modality_type: int)` method')
+
+            return self.generate_modality_only(batch_size = 1)
+
         # take care of prompt being a raw tensor, either text or raw modality (image, video, actions, latents, etc)
 
         if is_tensor(prompt) and prompt.dtype == torch.float: # is modality with type 0 implicit
@@ -1873,7 +1881,7 @@ class Transfusion(Module):
         # maybe add axial pos emb
 
         if mod.add_pos_emb:
-            axial_pos_emb = mod.pos_emb_mlp(tensor(axial_dims), flatten_dims = True)
+            axial_pos_emb = mod.pos_emb_mlp(tensor(axial_dims), flatten = True)
             noised_tokens = noised_tokens + axial_pos_emb
 
         # attention
@@ -2175,6 +2183,10 @@ class Transfusion(Module):
 
             return inner
 
+        # prepare storing of sizes of all modalities that require axial positions, for delayed application for efficiency
+
+        pos_emb_max_axial_dims: dict[int, list[Tensor]] = defaultdict(list)
+
         # go through all modality samples and do necessary transform
 
         for batch_index, batch_modalities in enumerate(modalities):
@@ -2329,9 +2341,9 @@ class Transfusion(Module):
                 if need_axial_pos_emb:
 
                     if exists(mod.pos_emb_mlp):
-                        pos_emb = mod.pos_emb_mlp(tensor(modality_shape_tuple), flatten_dims= True)
+                        pos_emb_max_axial_dims[modality_type].append(tensor(modality_shape_tuple))
+                        pos_emb = (modality_type, modality_shape_tuple, (precede_modality_tokens, succeed_modality_tokens))
 
-                        pos_emb = F.pad(pos_emb, (0, 0, precede_modality_tokens, succeed_modality_tokens), value = 0.)
                     else:
                         pos_emb = torch.zeros(text_tensor.shape[0], self.dim, device = device)
 
@@ -2340,7 +2352,7 @@ class Transfusion(Module):
             text.append(cat(batch_text))
 
             if need_axial_pos_emb:
-                modality_pos_emb.append(cat(batch_modality_pos_emb, dim = -2))
+                modality_pos_emb.append(batch_modality_pos_emb)
 
             modality_tokens.append(cat(batch_modality_tokens))
             modality_positions.append(batch_modality_positions)
@@ -2354,8 +2366,38 @@ class Transfusion(Module):
 
         modality_tokens = pad_sequence(modality_tokens, padding_value = 0.)
 
+        # handle modality positional embedding
+
         if need_axial_pos_emb:
-            modality_pos_emb = pad_sequence(modality_pos_emb, padding_value = 0.)
+            pos_emb_max_axial_dims = {mod_type: stack(sizes, dim = -1).amax(dim = -1) for mod_type, sizes in pos_emb_max_axial_dims.items()}
+            factorized_pos_emb = {mod_type: self.get_modality_info(mod_type).pos_emb_mlp(max_size, return_factorized = True) for mod_type, max_size in pos_emb_max_axial_dims.items()}
+
+            # lazy evaluate the modality positional embedding from the factorized positional embedding from maximum axial dims
+
+            evaluated_pos_emb = []
+
+            for batch_modality_pos_emb in modality_pos_emb:
+                evaluated_batch_pos_emb = []
+
+                for maybe_pos_emb_config in batch_modality_pos_emb:
+
+                    if is_tensor(maybe_pos_emb_config):
+                        evaluated_batch_pos_emb.append(maybe_pos_emb_config)
+                        continue
+
+                    mod_type, mod_size, padding = maybe_pos_emb_config
+
+                    mod_info = self.get_modality_info(mod_type)
+                    mod_factorized_pos_emb = factorized_pos_emb[mod_type]
+
+                    mod_pos_emb = mod_info.pos_emb_mlp.combine_factorized(mod_factorized_pos_emb, mod_size, flatten = True)
+                    mod_pos_emb = F.pad(mod_pos_emb, (0, 0, *padding), value = 0.) # handle padding for preceding and succeeding meta tokens
+
+                    evaluated_batch_pos_emb.append(mod_pos_emb)
+
+                evaluated_pos_emb.append(cat(evaluated_batch_pos_emb, dim = -2))
+
+            modality_pos_emb = pad_sequence(evaluated_pos_emb, padding_value = 0.)
 
         # handle training mode and removal of last token
 
